"""Main module."""

import boto3
import botocore.exceptions


class AwsIpLiberator:
    """AWS IP Liberator

    :param access_key:  AWS access key
    :param secret_key:  AWS secret access key
    :param region_name: AWS region name
    """

    def __init__(self, access_key: str, secret_key: str, region_name: str, tag: str = None):
        session = boto3.session.Session(
            aws_access_key_id=access_key,
            aws_secret_access_key=secret_key,
            region_name=region_name
        )
        self.ec2 = session.client('ec2')
        self.tag = tag

    def authorize_rule(self, rule: dict):
        # rule authorization
        try:
            self.ec2.authorize_security_group_ingress(**self.tagged_rule(rule))
        # in case of problems to authorize
        except botocore.exceptions.ClientError as e:
            # when a permission is duplicated
            if e.response['Error']['Code'] == 'InvalidPermission.Duplicate':
                ip_permissions = rule['IpPermissions']

                # single duplicated permission: results in a warning
                if len(ip_permissions) == 1:
                    description, ip = (ip_permissions[0]['IpRanges'][0][k]
                                       for k in ('Description', 'CidrIp'))
                    print("> Rule '%s' not set: permission with IP %s already exists" % (description, ip))
                # several permissions: retry authorize permission by permission
                else:
                    for ip_perm in ip_permissions:
                        self.authorize_rule({'GroupId': rule['GroupId'],
                                             'IpPermissions': [ip_perm]})
            else:
                raise e

    def revoke_rule(self, rule: dict):
        self.ec2.revoke_security_group_ingress(**rule)

    def describe_rules(self, services: dict, config: dict):
        if 'tag' in config:
            services = {'[%s] %s' % (config['tag'], key): value for key, value in services.items()}
        group_ids = config['security_groups']
        groups = self.ec2.describe_security_groups(GroupIds=group_ids)['SecurityGroups']

        for group in groups:
            group_id = group['GroupId']
            ip_permissions = group['IpPermissions'][:]

            for i in reversed(range(len(ip_permissions))):
                ip_perm = ip_permissions[i]
                ip_perm['IpRanges'] = [e for e in ip_perm['IpRanges']
                                       if 'Description' in e and e['Description'] in services]

                if not ip_perm['IpRanges']:
                    del ip_permissions[i]

            if not ip_permissions:
                continue

            yield {
                'GroupId': group_id,
                'IpPermissions': ip_permissions
            }

    def tagged_rule(self, rule: dict) -> dict:
        if not self.tag:
            return rule

        new_rule = {
            **rule,
            'IpPermissions': [
                {
                    **ip_permission,
                    'IpRanges': [
                        {
                            **ip_range,
                            'Description': self.tag_description(ip_range['Description'])
                        }
                        for ip_range in ip_permission['IpRanges']]
                }
                for ip_permission in rule['IpPermissions']
            ]
        }
        return new_rule

    def tag_description(self, description: str):
        return description if not self.tag else '[%s] %s' % (self.tag, description)



import argparse
import json
import sys
import os.path

from . import __version__
from .ip_liberator import AwsIpLiberator
from .utils import whats_my_ip, make_rules, duplicate_removal


def main(program=sys.argv[0], args=sys.argv[1:]):
    parser = argparse.ArgumentParser(prog=os.path.basename(program), description='AWS IP Liberator')
    parser.add_argument('--version', action='version', version='%(prog)s ' + __version__)
    parser.add_argument('-p', '--profile',
                        dest='settings', required=True, type=open, metavar='FILE',
                        help='Profile settings in JSON format')
    parser.add_argument('--operator',
                        dest='operator', required=False, default=None,
                        help='Override the operator informed in the config')
    parser.add_argument('--my-ip',
                        dest='my_ip', required=False, default=None,
                        help='Use this IP instead of discover current')
    parser.add_argument('-t', '--tag',
                        dest='tag', default='ip-liberator',
                        help='Help to identify entries added by IP Liberator on security group (default: %(default)s)')
    parser.add_argument('--no-tag', dest='tag', action='store_const', const=None, help=argparse.SUPPRESS)
    parser.add_argument('--revoke-only',
                        dest='revoke_only', action='store_true',
                        help='Only revoke the rules specified in the config')

    args = parser.parse_args(args)
    settings = json.load(args.settings)
    revoke_only = args.revoke_only

    if args.my_ip:
        whats_my_ip(args.my_ip)

    if args.operator:
        settings['config']['operator'] = args.operator

    if args.tag:
        print("[%s] tag used" % args.tag)
        settings['config']['tag'] = args.tag

    access_key = settings['credentials']['access_key']
    secret_key = settings['credentials']['secret_key']
    region_name = settings['credentials']['region_name']

    liberator = AwsIpLiberator(access_key, secret_key, region_name, tag=args.tag)

    # create index of services
    services = make_services_index(settings)

    # rules to revoke matching config entries
    revoking_rules = liberator.describe_rules(services, settings['config'])

    # don't authorize
    if revoke_only:
        print("Revoking rules", [svc for svc in services])
        for rule_to_revoke in revoking_rules:
            print('-', rule_to_revoke['GroupId'])
            liberator.revoke_rule(rule_to_revoke)

        return 0

    # when authorizing, make a index of the revoking rules
    revoking_rules = {rule['GroupId']: rule for rule in revoking_rules}

    # rules to authorize from groups and services in the config
    liberator_rules = make_rules(services, settings['config'])

    # use first rule to print ip addresses
    rule_to_authorize = next(liberator_rules)
    ip_addresses = set(ip_range['CidrIp']
                       for ip_permission in rule_to_authorize['IpPermissions']
                       for ip_range in ip_permission['IpRanges'])

    if len(ip_addresses) == 1:
        print("Authorizing rules", [svc for svc in services], "to IP", ip_addresses.pop())
    else:
        for ip_range in (r for p in rule_to_authorize['IpPermissions'] for r in p['IpRanges']):
            print("Authorizing rule '%s' to IP %s" % (ip_range['Description'], ip_range['CidrIp']))

    while True:
        group_id = rule_to_authorize['GroupId']
        print('-', group_id)

        # revoke rules in this security group if any
        rule_to_revoke = revoking_rules.get(group_id)

        if rule_to_revoke:
            duplicate_removal(rule_to_authorize, rule_to_revoke)

            if rule_to_revoke['IpPermissions']:
                liberator.revoke_rule(rule_to_revoke)

        # authorize rules with new ip
        if rule_to_authorize['IpPermissions']:
            liberator.authorize_rule(rule_to_authorize)

        try:
            rule_to_authorize = next(liberator_rules)
        except StopIteration:
            break

    return 0


def make_services_index(settings):
    operator = settings['config']['operator']
    return {'%s %s' % (operator, svc['name']): svc for svc in settings['config']['services']}



import re
import urllib.request

from typing import Optional, Iterator


def whats_my_ip(my_ip: str = None) -> Optional[str]:
    # change cached value and return
    if my_ip:
        whats_my_ip.cache = my_ip
        return my_ip

    # try to return cached value
    try:
        return whats_my_ip.cache
    except AttributeError:
        pass

    with urllib.request.urlopen("https://checkip.amazonaws.com/") as response:
        if response.status == 200:
            data = response.read().strip()
            my_ip = "%s/32" % data.decode()
            return whats_my_ip(my_ip)

    return None


def make_keys(rule: dict) -> dict:
    return {(ip_permission['FromPort'], ip_permission['ToPort'], ip_range['Description'], ip_range['CidrIp']): i
            for i, ip_permission in enumerate(rule['IpPermissions'])
            for ip_range in ip_permission['IpRanges']}


def duplicate_removal(rule1: dict, rule2: dict) -> None:
    rule_keys1 = make_keys(rule1)
    rule_keys2 = make_keys(rule2)

    indexes = [(rule_keys1[key], rule_keys2[key]) for key in rule_keys2 if key in rule_keys1]
    indexes = zip(sorted([t[0] for t in indexes], reverse=True),
                  sorted([t[1] for t in indexes], reverse=True))

    for i, j in indexes:
        del rule1['IpPermissions'][i]
        del rule2['IpPermissions'][j]


def make_rule(description: str, cidr_ip: str,
              *, port: int = None, from_port: int = None, to_port: int = None) -> dict:
    if port is not None:
        assert from_port is None and to_port is None, "cannot mix 'port' with 'from_port' and 'to_port'"
        from_port = to_port = port
    else:
        assert from_port is not None and to_port is not None, "port range not set"

    return {
        'FromPort': from_port,
        'ToPort': to_port,
        'IpProtocol': 'tcp',
        'IpRanges': [
            {
                'CidrIp': cidr_ip,
                'Description': description
            }
        ]
    }


RE_PORT_RANGE = re.compile(r"^[0-9]+(-[0-9]+)?$")


def make_rules(services: dict, config: dict) -> Iterator[dict]:
    group_ids = config['security_groups']

    ip_permissions = []
    for description, service in services.items():
        ip = service.get('ip') or whats_my_ip()
        port_spec = service['port'].strip()

        if not port_spec:
            raise ValueError("No port range informed in service: %s" % description)

        if not RE_PORT_RANGE.match(port_spec):
            raise ValueError("Invalid port range: '%s'" % port_spec)

        port_range = tuple(map(int, port_spec.split('-')))
        if len(port_range) == 1:
            port_range *= 2

        ip_permissions += [make_rule(description, ip, from_port=port_range[0], to_port=port_range[1])]

    for gid in group_ids:
        yield {
            'GroupId': gid,
            'IpPermissions': ip_permissions[:]
        }



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['ip_liberator']

package_data = \
{'': ['*']}

install_requires = \
['boto3==1.9.133']

entry_points = \
{'console_scripts': ['ip-liberator = ip_liberator.__main__:main']}

setup_kwargs = {
    'name': 'ip-liberator',
    'version': '0.2.3',
    'description': 'A command line utility to update AWS Security Groups rules.',
    'long_description': '============\nIP Liberator\n============\n\n\n.. image:: https://img.shields.io/pypi/v/ip-liberator.svg\n        :target: https://pypi.python.org/pypi/ip-liberator\n\n.. image:: https://img.shields.io/travis/wagnerluis1982/ip-liberator.svg\n        :target: https://travis-ci.org/wagnerluis1982/ip-liberator\n\n.. image:: https://readthedocs.org/projects/ip-liberator/badge/?version=latest\n        :target: https://ip-liberator.readthedocs.io/en/latest/?badge=latest\n        :alt: Documentation Status\n\n\nA command line utility to update AWS Security Groups rules.\n\n\n* Free software: GNU General Public License v3\n* Documentation: https://ip-liberator.readthedocs.io.\n\n\nFeatures\n--------\n\n* Update a list of security groups of your AWS account at once.\n* Grant access to informed ports for your current IP address or an informed IP.\n* Read profile files in JSON with all the information needed to contact.\n* Fit for use as script (e.g. to update your dynamic IP regularly).\n\nInstallation\n------------\n\n.. code-block:: console\n\n    $ pip install ip-liberator\n\nQuickstart\n----------\n\nConsider a file ``/path/my-profile.json`` with the following content:\n\n.. code-block:: json\n\n    {\n      "credentials": {\n        "access_key": "<AWS_ACCESS_KEY>",\n        "secret_key": "<AWS_SECRET_KEY>",\n        "region_name": "<AWS REGION>"\n      },\n      "config": {\n        "operator": "John",\n        "services": [\n          {\n            "name": "FTP+SFTP",\n            "port": "21-22"\n          },\n          {\n            "name": "HTTPS",\n            "port": "443"\n          }\n        ],\n        "security_groups": [\n          "sg-<GROUP_ID_1>",\n          "sg-<GROUP_ID_2>"\n        ]\n      }\n    }\n\nUsing the profile defined above will create or update two entries in the informed security groups:\n\n- **John FTP+SFTP** granting access for the current IP the ports 21 and 22.\n- **John HTTPS** granting access for the current IP the port 443.\n\nTo accomplish it, simply run:\n\n.. code-block:: console\n\n    $ ip-liberator --profile /path/my-profile.json\n    Authorizing rules [\'John FTP+SSH\', \'John HTTPS\'] to IP 192.30.253.112/32\n    - sg-<GROUP_ID_1>\n    - sg-<GROUP_ID_2>\n\nCredits\n-------\n\nAuthors\n:::::::\n\n* Wagner Macedo <wagnerluis1982@gmail.com> (maintainer)\n',
    'author': 'Wagner Macedo',
    'author_email': 'wagnerluis1982@gmail.com',
    'url': 'https://github.com/wagnerluis1982/ip-liberator',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
